from __future__ import print_function
from builtins import (super, zip)
from past.builtins import basestring

import textwrap
from copy import deepcopy

from . import core


class Constructs(core.Constructs):
    '''A container for metadata constructs.

Calling a `Constructs` instance selects metadata constructs by
identity and is an alias for the `filter_by_identity` method.

**Examples:**

Select constructs that have a "standard_name" property of 'foo':

>>> d = c('foo')

.. versionadded:: 1.7.0

    ''' 
    def __call__(self, *identities):
        '''Select metadata constructs by identity.

Calling a `Constructs` instance selects metadata constructs by
identity and is an alias for the `filter_by_identity` method.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_identity`

:Parameters:

    identities: optional
        See `filter_by_identity` for details.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

>>> c('latitude')
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>}
>>> c.filter_by_identity('latitude')
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>}

 See `filter_by_identity` for more examples.

        '''
        return self.filter_by_identity(*identities)
    #--- End: def
    
    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

.. versionadded:: 1.7.0

        '''
        construct_types = ['{0}({1})'.format(c, len(v))
                           for c, v in sorted(self._constructs.items())
                           if len(v) and c not in self._ignore]    
        
        return '<{0}: {1}>'.format(self.__class__.__name__, ', '.join(construct_types))
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        out = ['Constructs:']

        construct_types = [c
                           for c, v in sorted(self._constructs.items())
                           if len(v) and c not in self._ignore]    

        first = '{'
        for construct_type in construct_types:
            for key, value in sorted(self._constructs[construct_type].items()):
                if first:
                    out[0] = out[0] + '\n{{{!r}: {!r},'.format(key, value)
                    first = False
                else:
                    out.append('{!r}: {!r},'.format(key, value))
        #--- End: for
        
        if first:
            out[0] = out[0] + '\n{}'
        else:
            out[-1] = out[-1][:-1] + '}'

        return '\n '.join(out)
    #--- End: def

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _axes_to_constructs(self):
        '''Map domain axis constructs to the metadata constructs whose data
span them.

This is useful for ascertaining whether or not two `Constructs`
instances are equal.

.. versionadded:: 1.7.0

:Returns:

    `dict`

**Examples:**

>>> f.constructs._axes_to_constructs()
{('domainaxis0',): {'auxiliary_coordinate': {},
                    'cell_measure'        : {},
                    'dimension_coordinate': {'dimensioncoordinate0': <DimensionCoordinate: atmosphere_hybrid_height_coordinate(1) >},
                    'domain_ancillary'    : {'domainancillary0': <DomainAncillary: ncvar%a(1) m>,
                                             'domainancillary1': <DomainAncillary: ncvar%b(1) >},
                    'field_ancillary'     : {}},
 ('domainaxis1',): {'auxiliary_coordinate': {'auxiliarycoordinate2': <AuxiliaryCoordinate: long_name=Grid latitude name(10) >},
                    'cell_measure'        : {},
                    'dimension_coordinate': {'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>},
                    'domain_ancillary'    : {},
                    'field_ancillary'     : {}},
 ('domainaxis1', 'domainaxis2'): {'auxiliary_coordinate': {'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>},
                                  'cell_measure'        : {},
                                  'dimension_coordinate': {},
                                  'domain_ancillary'    : {'domainancillary2': <DomainAncillary: surface_altitude(10, 9) m>},
                                  'field_ancillary'     : {'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}},
 ('domainaxis2',): {'auxiliary_coordinate': {},
                    'cell_measure'        : {},
                    'dimension_coordinate': {'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>},
                    'domain_ancillary'    : {},
                    'field_ancillary'     : {}},
 ('domainaxis2', 'domainaxis1'): {'auxiliary_coordinate': {'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>},
                                  'cell_measure'        : {'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>},
                                  'dimension_coordinate': {},
                                  'domain_ancillary'    : {},
                                  'field_ancillary'     : {}},
 ('domainaxis3',): {'auxiliary_coordinate': {},
                    'cell_measure'        : {},
                    'dimension_coordinate': {'dimensioncoordinate3': <DimensionCoordinate: time(1) days since 2018-12-01 >},
                    'domain_ancillary'    : {},
                    'field_ancillary'     : {}}}

        '''
        out = {}


        data_axes = self.data_axes()
        
        for axes in data_axes.values():
            d = {construct_type: {}
                 for construct_type in self._array_constructs
                 if construct_type not in self._ignore}

            out[axes] = d
        #--- End: for

        for cid, construct in self.filter_by_data().items():
            axes = data_axes.get(cid)
            construct_type = self._construct_type[cid]
            out[axes][construct_type][cid] = construct

        return out
    #--- End: def

    def _equals_cell_method(self, other, rtol=None, atol=None,
                            verbose=False, ignore_type=False,
                            axis1_to_axis0=None, key1_to_key0=None):
        '''

        '''
        cell_methods0 = self.filter_by_type('cell_method')
        cell_methods1 = other.filter_by_type('cell_method')

        if len(cell_methods0) != len(cell_methods1):
            if verbose:
                print("Verbose: Different numbers of cell methods: {0!r} != {1!r}".format(
                    cell_methods0, cell_methods1))
            return False

        if not len(cell_methods0):
            return True    

        cell_methods0 = cell_methods0.ordered()
        cell_methods1 = cell_methods1.ordered()

        axis0_to_axis1 = {axis0: axis1
                          for axis1, axis0 in axis1_to_axis0.items()}

        for cm0, cm1 in zip(tuple(cell_methods0.values()),
                            tuple(cell_methods1.values())):
            # Check that there are the same number of axes
            axes0 = cm0.get_axes(())
            axes1 = list(cm1.get_axes(()))
            if len(axes0) != len(axes1):
                if verbose:
                    print("{0}: Different cell methods (mismatched axes):\n  {1}\n  {2}".format(
                        cm0.__class__.__name__, cell_methods0.ordered(), cell_methods1.ordered()))
                return False
    
            indices = []
            for axis0 in axes0:
                if axis0 is None:
                    return False
                for axis1 in axes1:
                    if axis0 in axis0_to_axis1 and axis1 in axis1_to_axis0:
                        if axis1 == axis0_to_axis1[axis0]:
                            # axis0 and axis1 are domain axis
                            # constructs that correspond to each other
                            axes1.remove(axis1)
                            indices.append(cm1.get_axes(()).index(axis1))
                            break
                    elif axis0 in axis0_to_axis1 or axis1 in axis1_to_axis0:
                        # Only one of axis0 and axis1 is a domain axis
                        # construct
                        if verbose:
                            print("{0}: Different cell methods (mismatched axes):\n  {1}\n  {2}".format(
                                cm0.__class__.__name__, cell_methods0, cell_methods1))
                        return False
                    elif axis0 == axis1:
                        # axes0 and axis 1 are identical standard
                        # names
                        axes1.remove(axis1)
                        indices.append(cm1.get_axes(()).index(axis1))
                    elif axis1 is None:
                        # axis1 
                        if verbose:
                            print("{0}: Different cell methods (mismatched axes):\n  {1}\n  {2}".format(
                                cm0.__class__.__name__, cell_methods0, cell_methods1))
                        return False
            #--- End: for

            if len(cm1.get_axes(())) != len(indices):
                if verbose:
                    print("{0}: [4] Different cell methods (mismatched axes):\n  {1}\n  {2}".format(
                        cm0.__class__.__name__, cell_methods0, cell_methods1))
                return False

            cm1 = cm1.sorted(indices=indices)
            cm1.set_axes(axes0)

            if not cm0.equals(cm1, atol=atol, rtol=rtol,
                              verbose=verbose,
                              ignore_type=ignore_type):
                if verbose:
                    print("Verbose: Different cell methods: {0!r}, {1!r}".format(
                        cell_methods0, cell_methods1))
                return False                
        #--- End: for

        return True
    #--- End: def

    def _equals_coordinate_reference(self, other, rtol=None,
                                     atol=None, verbose=0,
                                     ignore_type=False,
                                     axis1_to_axis0=None,
                                     key1_to_key0=None):
        '''TODO
        '''
        refs0 = dict(self.filter_by_type('coordinate_reference'))
        refs1 = dict(other.filter_by_type('coordinate_reference'))

        if len(refs0) != len(refs1):
            if verbose:
                print("{0}: Different numbers of {1} constructs: {2} != {3}".format(
                    self.__class__.__name__, 'coordinate reference',
                    len(refs0), len(refs1)))

            return False
        
        if refs0:
            construct_verbose = (verbose > 1)
        
            for ref0 in refs0.values():
                found_match = False
                for key1, ref1 in tuple(refs1.items()):
                    if construct_verbose:
                        print("{0}: Comparing {1!r}, {2!r}".format(
                            self.__class__.__name__, ref0, ref1), end=": ") # pragma: no cover
 
                    if not ref0.equals(ref1, rtol=rtol, atol=atol,
                                       verbose=construct_verbose,
                                       ignore_type=ignore_type):
                        continue

                    # Coordinates
                    coordinates0 = ref0.coordinates()
                    coordinates1 = set()
                    for value in ref1.coordinates():
                        coordinates1.add(key1_to_key0.get(value, value))
                        
                    if coordinates0 != coordinates1:
                        if construct_verbose:
                            print("Coordinates don't match") # pragma: no cover
                            
                        continue
    
                    # Domain ancillary-valued coordinate conversion terms
                    terms0 = ref0.coordinate_conversion.domain_ancillaries()

                    terms1 = {term: key1_to_key0.get(key, key)
                              for term, key in ref1.coordinate_conversion.domain_ancillaries().items()}

                    if terms0 != terms1:
                        if construct_verbose:
                            print("Coordinate conversion domain ancillaries don't match") # pragma: no cover
 
                        continue

                    if construct_verbose:
                        print("OK") # pragma: no cover
                        
                    found_match = True
                    del refs1[key1]                                       
                    break
                #--- End: for
    
                if not found_match:
                    if verbose:
                        print("{0}: No match for {1!r})".format(
                            self.__class__.__name__, ref0))
                    return False
            #--- End: for
        #--- End: if

        return True
    #--- End: def

    def _equals_domain_axis(self, other, rtol=None, atol=None,
                            verbose=False, ignore_type=False,
                            axis1_to_axis0=None, key1_to_key0=None):
        '''
        '''
        self_sizes  = [d.get_size()
                       for d in self.filter_by_type('domain_axis').values()]
        other_sizes = [d.get_size()
                       for d in other.filter_by_type('domain_axis').values()]
      
        if sorted(self_sizes) != sorted(other_sizes):
            # There is not a 1-1 correspondence between axis sizes
            if verbose:
                print("{0}: Different domain axis sizes: {1} != {2}".format(
                    self.__class__.__name__,
                    sorted(self_sizes), sorted(other_sizes)))
            return False

        return True
    #--- End: def
    
    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self, data=True):
        '''Return a deep copy.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

.. versionadded:: 1.7.0

:Parameters:

    data: `bool`, optional
        If False then do not copy data contained in the metadata
        constructs. By default such data are copied.

:Returns:

        The deep copy.

**Examples:**

>>> g = f.copy()
>>> g = f.copy(data=False)

        '''
        out = super().copy(data=data)

        prefiltered = getattr(self, '_prefiltered', None)
        if prefiltered is not None:
            out._prefiltered = prefiltered.copy(data=data)
            out._filters_applied = self._filters_applied
            
        return out
    #--- End: def

#    def domain_axis_key(self, identity, default=ValueError()):
#        '''Return the key of the domain axis construct that is spanned by 1-d
#coordinate constructs.
#
#.. versionadded:: 1.7.8
#
#.. seealso:: `domain_axis_identity`
#
#:Parameters:
#
#    identity:
#
#        Select the 1-d coordinate constructs that have the given
#        identity.
#
#        An identity is specified by a string (e.g. ``'latitude'``,
#        ``'long_name=time'``, etc.); or a compiled regular expression
#        (e.g. ``re.compile('^atmosphere')``), for which all constructs
#        whose identities match (via `re.search`) are selected.
#
#        Each coordinate construct has a number of identities, and is
#        selected if any of them match any of those provided. A
#        construct's identities are those returned by its `!identities`
#        method. In the following example, the construct ``x`` has four
#        identities:
#
#           >>> x.identities()
#           ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']
#
#        In addition, each construct also has an identity based its
#        construct key (e.g. ``'key%dimensioncoordinate2'``)
#
#        Note that in the output of a `print` call or `!dump` method, a
#        construct is always described by one of its identities, and so
#        this description may always be used as an *identity* argument.
#
#    default: optional
#        Return the value of the *default* parameter if a domain axis
#        construct can not be found. If set to an `Exception` instance
#        then it will be raised instead.
#
#:Returns:
#
#    `str`
#        The key of the domain axis construct that is spanned by the
#        data of the selected 1-d coordinate constructs.
#
#**Examples:**
#
#TODO
#
#        '''
#        # Select 1-d coordinate constructs with the given identity
#        c = self.filter_by_type('dimension_coordinate',
#                                'auxiliary_coordinates')
#        c = c.filter_by_naxes(1)
#        c = c.filter_by_identity(identity)
#        
#        if not len(c) :
#            return self._default(
#                default,
#                "No 1-d coordinate constructs have identity {!r}".format(identity))
#
#        data_axes = self.data_axes()
#        domain_axes = self.filter_by_type('domain_axis')
#        
#        keys = []
#        for ckey, coord in c.items():
#            axes = data_axes.get(ckey)
#            if not axes:
#                continue
#            
#            key = axes[0]
#            if domain_axes.get(key):
#                keys.append(key)
#        #--- End: for
#        
#        keys = set(keys)
#            
#        if not keys:
#            return self._default(
#                default,
#                "1-d coordinate constructs selected with identity {!r} have not been assigned a domain axis contructs".format(
#                    coord))                       
#        
#        if len(keys) > 1:
#            return self._default(
#                default,
#                "Multiple 1-d coordinate constructs selected with identity {!r} span multiple domain axes: {!r}".format(
#                    identity, keys))
#
#        return keys.pop()
#    #--- End: def

    def domain_axis_identity(self, key):
        '''Return the canonical identity for a domain axis construct.

The identity is the first found of the following:

1. The canonical identity of a dimension coordinate contruct that span
   the domain axis construct.
2. The identity of a one-dimensional auxiliary coordinate construct
   that spans the domain axis construct. This will either be the value
   of a "standard_name", "cf_role" (preceeded by ``'cf_role='``) or
   "axis" (preceeded by ``'axis='``) property, as appropriate.
3. The netCDF dimension name, preceeded by 'ncdim%'.
4. The domain axis construct key, preceeded by 'key%'.

:Parameters:

    key: `str`
        The construct key for the domain axis construct.

        *Parameter example:*
          ``key='domainaxis2'``

:Returns:

    `str`
        The identity.

**Examples:**

>>> c.domain_axis_identity('domainaxis1')
'longitude'
>>> c.domain_axis_identity('domainaxis2')
'axis=Y'
>>> c.domain_axis_identity('domainaxis3')
'cf_role=timeseries_id'
>>> c.domain_axis_identity('domainaxis4')
'key%domainaxis4')

        '''
        domain_axes = self.filter_by_type('domain_axis')
        
        if key not in domain_axes:
            raise ValueError(
                'No domain axis construct with key {!r}'.format(key))

        constructs_data_axes = self.data_axes()

        # Try to get the identity from an dimension coordinate
        identity = ''
        for dkey, dim in self.filter_by_type('dimension_coordinate').items():
            if constructs_data_axes.get(dkey) == (key,):
                identity = dim.identity()
                if identity.startswith('ncvar%'):
                    identity = ''
                
                break
        #--- End: for
        if identity:
            return identity

        # Try to get the identity from an auxiliary coordinate
        identities = []
        for akey, aux in self.filter_by_type('auxiliary_coordinate').items():
            if constructs_data_axes.get(akey) == (key,):
                identity = aux.identity()
                if not identity.startswith('ncvar%'):
                    identities.append(identity)
        #--- End: for

        if len(identities) == 1:
            return identities[0]
        elif len(identities) > 1:
            cf_role = []
            axis = []
            for i in identities:
                if i.startswith('axis='):
                    axis.append(i)
                elif i.startswith('cf_role='):
                    cf_role.append(i)
            #--- End: for
            
            if len(cf_role) == 1:
                return cf_role[0]
            
            if len(axis) == 1:
                return axis[0]
        #--- End: if

        # Try to get the identity from an netCDF dimension name
        ncdim = domain_axes[key].nc_get_dimension(None)
        if ncdim is not None:
            return 'ncdim%{0}'.format(ncdim)

        # Get the identity from the domain axis construct key
        return 'key%{0}'.format(key)
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_compression=False, _ignore_type=False,
               _return_axis_map=False):
        '''Whether two `Constructs` instances are the same.

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

If data arrays are compressed then the compression type and the
underlying compressed arrays must be the same, as well as the arrays
in their uncompressed forms. See the *ignore_compression* parameter.

Any type of object may be tested but equality is only possible with
another `Constructs` construct, or a subclass of one.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked on any
construct.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the "_FillValue" and "missing_value" properties
        are omitted from the comparison for the metadata constructs.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_compression: `bool`, optional
        If True then any compression applied to underlying arrays is
        ignored and only uncompressed arrays are tested for
        equality. By default the compression type and, if applicable,
        the underlying compressed arrays must be the same, as well as
        the arrays in their uncompressed forms

:Returns: 
  
    `bool`
        Whether the two instances are equal.

**Examples:**

>>> x.equals(x)
True
>>> x.equals(x.copy())
True
>>> x.equals('something else')
False

        '''        
        if self is other:
            if not _return_axis_map:
                return True

        construct_verbose = (verbose > 1)
        
        # Check that each instance is the same type
        if  not isinstance(other, self.__class__):
            if verbose:
                print("{0}: Incompatible type: {1}".format(
                    self.__class__.__name__, other.__class__.__name__))
            if not _return_axis_map:
                return False
        
        axes0_to_axes1 = {}
        axis0_to_axis1 = {}
        axis1_to_axis0 = {}
        key1_to_key0   = {}
        
        # ------------------------------------------------------------
        # Domain axis constructs
        # ------------------------------------------------------------
        if not self._equals_domain_axis(other, rtol=rtol, atol=atol,
                                        verbose=verbose,
                                        ignore_type=_ignore_type,
                                        axis1_to_axis0=axis1_to_axis0,
                                        key1_to_key0=key1_to_key0):
            if not _return_axis_map:
                return False
        
        # ------------------------------------------------------------
        # Constructs with arrays
        # ------------------------------------------------------------
        log = []
        axes_to_constructs0 = self._axes_to_constructs()
        axes_to_constructs1 = other._axes_to_constructs()
        for axes0, constructs0 in axes_to_constructs0.items():
            matched_all_constructs_with_these_axes = False

             
            len_axes0 = len(axes0) 
            for axes1, constructs1 in tuple(axes_to_constructs1.items()):

                constructs1 = constructs1.copy()

                if len_axes0 != len(axes1):
                    # axes1 and axes0 contain different number of
                    # domain axes.
                    continue

                for construct_type in self._array_constructs:
#                    role_constructs0 = constructs0[construct_type]
#                    role_constructs1 = constructs1[construct_type].copy()
                    role_constructs0 = constructs0.get(construct_type, {})
                    role_constructs1 = constructs1.get(construct_type, {}).copy()
                    
                    if len(role_constructs0) != len(role_constructs1):
                        # There are the different numbers of
                        # constructs of this type
                        matched_all_constructs_with_these_axes = False
                        log.append("{0}: Different numbers of {1} constructs: {2} != {3}".format(
                            self.__class__.__name__, construct_type,
                            len(role_constructs0), len(role_constructs1)))
                        break
#                    elif not role_constructs0:
#                        break

                    # Check that there are matching pairs of equal
                    # constructs
                    matched_construct = True
                    for key0, item0 in role_constructs0.items():
                        matched_construct = False
                        for key1, item1 in tuple(role_constructs1.items()):
                            if construct_verbose:
                                print("{}: Comparing {!r}, {!r}".format(
                                    self.__class__.__name__, item0, item1), end=": ") # pragma: no cover
                                
                            if item0.equals(item1,
                                            rtol=rtol, atol=atol,
                                            verbose=construct_verbose,
                                            ignore_data_type=ignore_data_type,
                                            ignore_fill_value=ignore_fill_value,
                                            ignore_compression=ignore_compression,
                                            ignore_type=_ignore_type):
                                if construct_verbose:
                                    print("OK") # pragma: no cover
                                    
                                del role_constructs1[key1]
                                key1_to_key0[key1] = key0
                                matched_construct = True
                                break
                        #--- End: for

                        if not matched_construct:
                            log.append("{0}: Can't match {1!r}".format(
                                self.__class__.__name__, item0))
                            break
                    #--- End: for

                    if role_constructs1:
                        # At least one construct in other is not equal
                        # to a construct in self
                        break

                    # Still here? Then all constructs of this type
                    # that spanning these axes match
#                    del constructs1[construct_type]
                    constructs1.pop(construct_type, None)
                #--- End: for

                matched_all_constructs_with_these_axes = not constructs1
                if matched_all_constructs_with_these_axes:
                    del axes_to_constructs1[axes1]
                    break
            #--- End: for

            if not matched_all_constructs_with_these_axes:
                if verbose:
                    names = [self.domain_axis_identity(axis0) for axis0 in axes0]
                    print("{0}: Can't match constructs spanning axes {1}".format(
                        self.__class__.__name__, names))
                    if log:
                        print('\n'.join(log))
                if not _return_axis_map:
                    return False
            else:
                # Map item axes in the two instances
                axes0_to_axes1[axes0] = axes1
        #--- End: for

        for axes0, axes1 in axes0_to_axes1.items():
            for axis0, axis1 in zip(axes0, axes1):
                if axis0 in axis0_to_axis1 and axis1 != axis0_to_axis1[axis0]:
                    if verbose:
                        print(
                            "{0}: Ambiguous axis mapping ({1} -> both {2} and {3})".format(
                                self.__class__.__name__,
                                self.domain_axis_identity(axes0),
                                other.domain_axis_identity(axis1),
                                other.domain_axis_identity(axis0_to_axis1[axis0]))) # pragma: no cover
                    if not _return_axis_map:
                        return False
                elif axis1 in axis1_to_axis0 and axis0 != axis1_to_axis0[axis1]:
                    if verbose:
                        print(
                            "{0}: Ambiguous axis mapping ({1} -> both {2} and {3})".format(
                                self.__class__.__name__,
                                self.domain_axis_identity(axis0),
                                self.domain_axis_identity(axis1_to_axis0[axis0]),
                                other.domain_axis_identity(axes1))) # pragma: no cover
                    if not _return_axis_map:
                        return False

                axis0_to_axis1[axis0] = axis1
                axis1_to_axis0[axis1] = axis0
        #--- End: for

        if _return_axis_map:
            return axis0_to_axis1
        
        # ------------------------------------------------------------
        # Constructs with no arrays
        # ------------------------------------------------------------
        for construct_type in self._non_array_constructs:
            if not getattr(self, '_equals_'+construct_type)(
                    other,
                    rtol=rtol, atol=atol, verbose=verbose,
                    ignore_type=_ignore_type,
                    axis1_to_axis0=axis1_to_axis0,
                    key1_to_key0=key1_to_key0):
                return False

        # ------------------------------------------------------------
        # Still here? Then the two objects are equal
        # ------------------------------------------------------------     
        return True
    #--- End: def

    def filter_by_axis(self, mode=None, *axes):
        '''Select metadata constructs by axes spanned by their data.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_data`, `filter_by_key`, `filter_by_measure`,
             `filter_by_method`, `filter_by_identity`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    mode: `str`
        Define the relationship between the given domain axes and the
        constructs' data.

        ==========  ==================================================
        *mode*      Description
        ==========  ==================================================
        ``'and'``   A construct is selected if it spans *all* of the 
                    given domain axes, *and possibly others*.

        ``'or'``    A construct is selected if it spans *any* of the
                    domain axes, *and possibly others*.
       
        ``exact``   A construct is selected if it spans *all* of the 
                    given domain axes, *and no others*.
       
        ``subset``  A construct is selected if it spans *a subset* of
                    the given domain axes, *and no others*.
        ==========  ==================================================

    axes: optional
        Select the constructs whose data spans particular domain axis
        constructs.

        A domain axis construct is identified by its construct key
        (e.g. ``'domainaxis1'``).

        If no axes are provided then all constructs that do or could
        have data, spanning any domain axes constructs, are selected.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select constructs whose data spans the "domainaxis1" domain axis
construct:

>>> d = c.filter_by_axis('and', 'domainaxis1')

Select constructs whose data does not span the "domainaxis2" domain
axis construct:

>>> d = c.filter_by_axis('and', 'domainaxis2').inverse_filter()

Select constructs whose data spans the "domainaxis1", but not the
"domainaxis2" domain axis constructs:

>>> d = c.filter_by_axis('and', 'domainaxis1')
>>> d = d.filter_by_axis('and', 'domainaxis2')
>>> d  = d.inverse_filter(1)

Select constructs whose data spans the "domainaxis1" or the
"domainaxis2" domain axis constructs:

>>> d = c.filter_by_axis('or', 'domainaxis1', 'domainaxis2')

        '''       
        out = self.shallow_copy()

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_axis': (mode, axes)},)
        
        # Parse the mode parameter
        _and      = False
        _or       = False
        _exact    = False
        _subset   = False
#        _superset = False

        if not axes and mode is None:
            mode = 'and'
        
        if mode == 'and':
            _and = True
        elif mode == 'or':
            _or = True
        elif mode == 'and':
            _and = True
        elif mode == 'exact':
            _exact = True
        elif mode == 'subset':
            _subset = True
 #       elif mode == 'superset':
 #           _superset = True
        else:    
            raise ValueError(
                "mode parameter must be one of 'and', 'or', 'exact', subset'")
        
        data_contructs = self.filter_by_data()
        constructs_data_axes = self.data_axes()

        axes = set(axes)
                    
        if not axes:            
            for cid in tuple(out):
                if cid not in data_contructs:
                    out._pop(cid)
            #--- End: for
            
            return out
        
        # Still here?
        for cid in tuple(out):
            if cid not in data_contructs:
                out._pop(cid)
                continue
                
            x = constructs_data_axes.get(cid)
            if x is None:
                # This construct does not have data axes
                out._pop(cid)
                continue
            
            ok = True
            if _exact:
                if set(x) != axes:
                    ok = False
            elif _subset:
                if not set(x).issubset(axes):
                    ok = False
#            elif _superset:
#                if not set(x).issuperset(axes):
#                    ok = False
            else:
                for axis_key in axes:
                    ok = axis_key in x
                    if _or:
                        if ok:
                            break
                    elif not ok:
                        break
            #--- End: if

            if not ok:
                # This construct ..
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_data(self):
        '''Select metadata constructs by whether they could contain data.

Selection is not based on whether they thay actually have data, rather
by whether the construct supports the inclusion of data. For example,
constructs selected by this method will all have a `!get_data` method.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_key`, `filter_by_measure`,
             `filter_by_method`, `filter_by_identity`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select constructs that could contain data:

>>> d = c.filter_by_data()

        '''       
        out = self.shallow_copy()

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_data': ()},)
        
        for cid in tuple(out):
            if out._construct_type[cid] not in self._array_constructs:
                # This construct can not have data
                out._pop(cid)
        #--- End: for
                
        return out
    #--- End: def

    def filter_by_identity(self, *identities):
        '''Select metadata constructs by identity.

Calling a `Constructs` instance selects metadata constructs by
identity and is an alias for the `filter_by_identity` method.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    identities: optional

        Select constructs that have any of the given identities.

        An identity is specified by a string (e.g. ``'latitude'``,
        ``'long_name=time'``, etc.); or a compiled regular expression
        (e.g. ``re.compile('^atmosphere')``), for which all constructs
        whose identities match (via `re.search`) are selected.

        If no identities are provided then all constructs are selected.

        Each construct has a number of identities, and is selected if
        any of them match any of those provided. A construct's
        identities are those returned by its `!identities` method. In
        the following example, the construct ``x`` has four
        identities:

           >>> x.identities()
           ['time', 'long_name=Time', 'foo=bar', 'ncvar%T']

        In addition, each construct also has an identity based its
        construct key (e.g. ``'key%dimensioncoordinate2'``)

        Note that in the output of a `print` call or `!dump` method, a
        construct is always described by one of its identities, and so
        this description may always be used as an *identities* argument.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select constructs that have a "standard_name" property of 'latitude':

>>> d = c.filter_by_identity('latitude')

Select constructs that have a "long_name" property of 'Height':

>>> d = c.filter_by_identity('long_name=Height')

Select constructs that have a "standard_name" property of 'latitude'
or a "foo" property of 'bar':

>>> d = c.filter_by_identity('latitude', 'foo=bar')

Select constructs that have a netCDF variable name of 'time':

>>> d = c.filter_by_identity('ncvar%time')

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_identity': identities},)
        
        # Return all constructs if no identities have been provided
        if not identities:
            return out
        
        for cid, construct in tuple(out.items()):
            ok = False
            for value0 in identities:          
                for value1 in ['key%'+cid] + construct.identities():
                    ok = self._matching_values(value0, construct, value1)
                    if ok:
                        break
                #--- End: for

                if ok:
                    break
            #--- End: for

            if not ok:
                # This construct does not match any of the identities
                out._pop(cid)
        #--- End: for

        return out
    #--- End: def

    def filter_by_key(self, *keys):
        '''Select metadata constructs by key.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_measure`,
             `filter_by_method`, `filter_by_identity`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    keys: optional
        Select constructs that have any of the given construct keys.

        A key is specified by a string (e.g. ``'fieldancillary0'``).

        If no keys are provided then all constructs are selected.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select the construct with key 'domainancillary0':

>>> d = c.filter_by_key('domainancillary0')

Select the constructs with keys 'dimensioncoordinate1' or
'fieldancillary0':

>>> d = c.filter_by_key('dimensioncoordinate1', 'fieldancillary0')

        '''
        out = self.shallow_copy()

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_key': keys},)
        
        if not keys:
            return out
        
        for cid in tuple(out):
            if cid not in keys:
                out._pop(cid)
        #--- End: for

        return out
    #--- End: def

    def filter_by_measure(self, *measures):
        '''Select cell measure constructs by measure.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_method`, `filter_by_identity`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    measures: optional
        Select cell measure constructs that have any of the given
        measure values.

        A measure is specified by a string (e.g. ``'area'``); or a
        compiled regular expression (e.g. ``re.compile('^a')``), for
        which all constructs whose measures match (via `re.search`)
        are selected.

        If no measures are provided then all cell measure constructs
        are selected.

:Returns:

    `Constructs`
        The selected cell measure constructs and their construct
        keys.
        
**Examples:**

>>> print(t.constructs.filter_by_type('measure'))
Constructs:
{'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>,
 'cellmeasure1': <CellMeasure: measure:volume(3, 9, 10) m3>}

Select cell measure constructs that have a measure of 'area':

>>> print(c.filter_by_measure('area'))
Constructs:
{'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>}

Select cell measure constructs that have a measure of 'area' or
'volume':

>>> print(c.filter_by_measure('area', 'volume'))
Constructs:
{'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>,
 'cellmeasure1': <CellMeasure: measure:volume(3, 9, 10) m3>}

Select cell measure constructs that have a measure of start with the
letter "a" or "v":

>>> print(c.filter_by_measure(re.compile('^a|v')))
Constructs:
{'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>,
 'cellmeasure1': <CellMeasure: measure:volume(3, 9, 10) m3>}

Select cell measure constructs that have a measure of any value:

>>> print(c.filer_by_measure())
Constructs:
{'cellmeasure0': <CellMeasure: measure:area(9, 10) km2>,
 'cellmeasure1': <CellMeasure: measure:volume(3, 9, 10) m3>}

        '''
        out = self.shallow_copy()

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_measure': measures},)
        
        for cid, construct in tuple(out.items()):
            try:
                get_measure = construct.get_measure
            except AttributeError:
                # This construct doesn't have a "get_measure" method
                out._pop(cid)
                continue

            if not measures:
#                if not construct.has_measure():
#                    out._pop(cid)
                continue

            ok = False
            for value0 in measures:
                value1 = construct.get_measure(None)
                ok = self._matching_values(value0, construct, value1)
                if ok:
                    break
            #--- End: for
            
            if not ok:
                # This construct does not match any of the measures
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_method(self, *methods):
        '''Select cell method constructs by method.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_identity`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    methods: optional
        Select cell method constructs that have any of the given
        methods.

        A method is specified by a string (e.g. ``'mean'``); or a
        compiled regular expression (e.g. ``re.compile('^m')``), for
        which all constructs whose methods match (via `re.search`) are
        selected.

        If no methods are provided then all cell method constructs are
        selected.

:Returns:

    `Constructs`
        The selected cell method constructs and their construct keys.

**Examples:**

>>> print(c.constructs.filter_by_type('cell_method'))
Constructs:
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>}

Select cell method constructs that have a method of 'mean':

>>> print(c.filter_by_method('mean'))
Constructs:
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>}

Select cell method constructs that have a method of 'mean' or
'maximum':

>>> print(c.filter_by_method('mean', 'maximum'))
Constructs:
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>}

Select cell method constructs that have a method that contain the letter 'x':

>>> import re
>>> print(c.filter_by_method(re.compile('x')))
Constructs:
{'cellmethod1': <CellMethod: domainaxis3: maximum>}

Select cell method constructs that have a method of any value:

>>> print(c.filter_by_method())
Constructs:
{'cellmethod0': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod1': <CellMethod: domainaxis3: maximum>}

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_method': methods},)
        
        for cid, construct in tuple(out.items()):
            try:
                get_method = construct.get_method
            except AttributeError:
                # This construct doesn't have a "get_method" method
                out._pop(cid)
                continue

            if not methods:
#                if not construct.has_method():
#                    out._pop(cid)
                continue
            
            ok = False
            for value0 in methods:
                value1 = get_method(None)
                ok = self._matching_values(value0, construct, value1)
                if ok:
                    break
            #--- End: for
            
            if not ok:
                # This construct does not match any of the methods
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_naxes(self, *naxes):
        '''Select metadata constructs by the number of domain axis contructs
spanned by their data.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_identity`, `filter_by_ncdim`,
             `filter_by_ncvar`, `filter_by_property`,
             `filter_by_type`, `filters_applied`, `inverse_filter`,
             `unfilter`

:Parameters:

    naxes: optional
        Select constructs whose data spans a particular number of
        domain axis constructs.

        A number of domain axis constructs is given by an `int`.

        If no numbers are provided then all constructs that do or
        could have data, spanning any domain axes constructs, are
        selected.

:Returns:

    `Constructs`
        The selected domain axis constructs and their construct keys.

**Examples:**

Select constructs that contain data that spans two domain axis
constructs:

>>> d = c.filter_by_naxes(2)

Select constructs that contain data that spans one or two domain axis
constructs:

>>> d = c.filter_by_ncdim(1, 2)

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_naxes': naxes},)
            
        data_contructs = self.filter_by_data()
        constructs_data_axes = self.data_axes()
        
        for key in tuple(out):
            if key not in data_contructs:
                out._pop(key)
                continue

            x = constructs_data_axes.get(key)
            if x is None:
                # This construct does not have data axes
#                out._pop(key)
                continue

            ok = True
            for n in naxes:
                if n == len(x):
                    ok = True
                    break

                ok = False

            if not ok:
                # This construct does not have the right number of axes
                out._pop(key)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_ncdim(self, *ncdims):
        '''Select domain axis constructs by netCDF dimension name.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncvar`, `filter_by_property`,
             `filter_by_type`, `filters_applied`, `inverse_filter`,
             `unfilter`

:Parameters:

    ncdims: optional
        Select domain axis constructs that have any of the given
        netCDF dimension names.

        A netCDF dimension name is specified by a string
        (e.g. ``'time'``); or a compiled regular expression
        (e.g. ``re.compile('^lat')``), for which all constructs whose
        netCDF dimension names match (via `re.search`) are selected.

        If no netCDF dimension names are provided then all domain axis
        constructs are selected.

:Returns:

    `Constructs`
        The selected domain axis constructs and their construct keys.

**Examples:**

Select the domain axis constructs with netCDF dimension name 'time':

>>> d = c.filter_by_ncdim('time')

Select the domain axis constructs with netCDF dimension name 'time' or
'lat':

>>> d = c.filter_by_ncdim('time', 'lat')

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_ncdim': ncdims},)
        
        for cid, construct in tuple(out.items()):
            try:
                nc_get_dimension = construct.nc_get_dimension
            except AttributeError:
                # This construct doesn't have a "nc_get_dimension"
                # method
                out._pop(cid)
                continue
            
            if not ncdims:
#                if not construct.nc_has_dimension():
#                    out._pop(cid)
                    
                continue
            
            ok = False
            for value0 in ncdims:
                value1 = nc_get_dimension(None)
                ok = self._matching_values(value0, construct, value1)
                if ok:
                    break
            #--- End: for
            
            if not ok:
                # This construct does not match any of the netCDF
                # dimension names
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_ncvar(self, *ncvars):
        '''Select domain axis constructs by netCDF variable name.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_property`,
             `filter_by_type`, `filters_applied`, `inverse_filter`,
             `unfilter`

:Parameters:

    ncvars: optional
        Select constructs that have any of the given netCDF variable
        names.

        A netCDF variable name is specified by a string
        (e.g. ``'time'``); or a compiled regular expression
        (e.g. ``re.compile('^lat')``), for which all constructs whose
        netCDF variable names match (via `re.search`) are selected.

        If no netCDF variable names are provided then all constructs
        that do or could have a netCDF variable name, with any value,
        are selected.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select the constructs with netCDF variable name 'time':

>>> d = c.filter_by_ncvar('time')

Select the constructs with netCDF variable name 'time' or 'lat':

>>> d = c.filter_by_ncvar('time', 'lat')

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_ncvar': ncvars},)
        
        for cid, construct in tuple(out.items()):            
            try:
                nc_get_variable = construct.nc_get_variable
            except AttributeError:
                # This construct doesn't have a "nc_get_variable"
                # method
                out._pop(cid)
                continue

            if not ncvars:
#                if not construct.nc_has_variable():
#                    out._pop(cid)                   
                continue

            ok = False
            for value0 in ncvars:
                value1 = nc_get_variable(None)
                ok = self._matching_values(value0, construct, value1)
                if ok:
                    break
            #--- End: for

            if not ok:
                # This construct does not match any of the netCDF
                # variable names
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def _matching_values(self, value0, construct, value1):
        if value1 is not None:                        
            try:
                result = value0.search(value1)
            except (AttributeError, TypeError):
                result = construct._equals(value1, value0)
                
            if result:
                # This construct matches this property
                return True

        return False
    
    def filter_by_property(self, *mode, **properties):
        '''Select metadata constructs by property.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:
        
    mode: optional
        Define the behaviour when multiple properties are provided.

        By default (or if the *mode* parameter is ``'and'``) a
        construct is selected if it matches all of the given
        properties, but if the *mode* parameter is ``'or'`` then a
        construct will be selected when at least one of its properties
        matches.

    properties:  optional
        Select constructs that have properties with the given
        values.

        By default a construct is selected if it matches all of the
        given properties, but it may alternatively be selected when at
        least one of its properties matches (see the *mode* positional
        parameter).

        A property value is given by a keyword parameter of the
        property name. The value may be a scalar or vector
        (e.g. ``'latitude'``, ``4``, ``['foo', 'bar']``); or a
        compiled regular expression (e.g. ``re.compile('^ocean')``),
        for which all constructs whose methods match (via `re.search`)
        are selected.

        If no properties are provided then all constructs that do or
        could have properties, with any values, are selected.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select constructs that have a "standard_name" of 'latitude':

>>> d = c.filter_by_property(standard_name='latitude')

Select constructs that have a "long_name" of 'height' *and* "units" of
'm':

>>> d = c.filter_by_property(long_name='height', units='m')

Select constructs that have a "long_name" of 'height' *or* a "foo" of
'bar':

>>> d = c.filter_by_property('or', long_name='height', foo='bar')

Select constructs that have a "standard_name" which contains start
with the string 'air':

>>> import re
>>> d = c.filter_by_property(standard_name=re.compile('^air'))

        '''
        out = self.shallow_copy()

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_property': (mode, properties)},)
        
        _or = False
        if mode:
            if len(mode) > 1:
                raise ValueError("Can provide at most one positional argument")
            
            x = mode[0]
            if x == 'or':
                _or = True
            elif x != 'and':
                raise ValueError("Positional argument, if provided, must 'or' or 'and'")
        #--- End: if

        for cid, construct in tuple(out.items()):
            try:
                get_property = construct.get_property
            except AttributeError:
                # This construct doesn't have a "get_property" method
                out._pop(cid)
                continue

            if not properties:
#                if not construct.properties():
#                    out._pop(cid)                    
                continue

            ok = True
            for name, value0 in properties.items():
                value1 = get_property(name, None)
                ok = self._matching_values(value0, construct, value1)

                if _or:
                    if ok:
                        break
                elif not ok:
                    break
            #--- End: for
            
            if not ok:
                # This construct does not match any of the sets of
                # properties
                out._pop(cid)
        #--- End: for

        return out
    #--- End: def
    
    def filter_by_size(self, *sizes):
        '''Select domain axis constructs by size.

.. versionadded:: 1.7.3

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:
        
    sizes: optional
        Select domain axis constructs that have any of the given
        sizes.

        A size is specified by an `int`.

        If no sizes are provided then all domain axis constructs are
        selected.

:Returns:

    `Constructs`
        The selected domain axis constructs and their construct keys.
        
**Examples:**

Select domain axis constructs that have a size of 1:

>>> d = c.filter_by_size(1)

Select domain axis constructs that have a size of 1 or 96:

>>> d = c.filter_by_size(1, 96)

        '''
        out = self.shallow_copy()
        
        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_size': sizes},)
        
        for cid, construct in tuple(out.items()):            
            try:
                get_size = construct.get_size
            except AttributeError:
                # This construct doesn't have a "get_size" method
                out._pop(cid)
                continue

            if not sizes:
#                if not construct.has_size():
#                    out._pop(cid)
                continue

            ok = False
            value0 = construct.get_size(None)
            for value1 in sizes:
                ok = self._matching_values(value1, construct, value0)
                if ok:
                    break
            #--- End: for
            
            if not ok:
                # This construct does not match any of the sizes
                out._pop(cid)
        #--- End: for
        
        return out
    #--- End: def

    def filter_by_type(self, *types):
        '''Select metadata constructs by type.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_identity`, `filter_by_property`,
             `filters_applied`, `inverse_filter`, `unfilter`

:Parameters:

    types: optional 
        Select constructs that have are of any of the given types.

        A type is specified by one of the following strings:

          ==========================  ================================
          *type*                      Construct selected
          ==========================  ================================
          ``'domain_ancillary'``      Domain ancillary constructs
          ``'dimension_coordinate'``  Dimension coordinate constructs
          ``'domain_axis'``           Domain axis constructs
          ``'auxiliary_coordinate'``  Auxiliary coordinate constructs
          ``'cell_measure'``          Cell measure constructs
          ``'coordinate_reference'``  Coordinate reference constructs
          ``'cell_method'``           Cell method constructs
          ``'field_ancillary'``       Field ancillary constructs
          ==========================  ================================

        If no types are provided then all constructs are selected.

:Returns:

    `Constructs`
        The selected constructs and their construct keys.

**Examples:**

Select dimension coordinate constructs:

>>> d = c.filter_by_type('dimension_coordinate')

Select dimension coordinate and field ancillary constructs:

>>> d = c.filter_by_type('dimension_coordinate', 'field_ancillary')

        '''
        out = super().filter_by_type(*types)

        out._prefiltered = self.shallow_copy()
        out._filters_applied = self.filters_applied() + ({'filter_by_type': types},)
        
        return out        
    #--- End: def
    
    def filters_applied(self):
        '''A history of filters that have been applied.

The history is returned in a tuple. The last element of the tuple
describes the last filter applied. Each element is a single-entry
dictionary whose key is the name of the filter method that was used,
with a value that gives the arguments that were passed to the call of
that method. If no filters have been applied then the tuple is empty.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`, `inverse_filter`,
             `unfilter`

:Returns:

    `tuple`
        The history of filters that have been applied, ordered from
        first to last. If no filters have been applied then the tuple
        is empty.


**Examples:**

>>> print(c)
{'auxiliarycoordinate0': <AuxiliaryCoordinate: latitude(10, 9) degrees_N>,
 'auxiliarycoordinate1': <AuxiliaryCoordinate: longitude(9, 10) degrees_E>,
 'coordinatereference1': <CoordinateReference: grid_mapping_name:rotated_latitude_longitude>,
 'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>,
 'dimensioncoordinate2': <DimensionCoordinate: grid_longitude(9) degrees>,
 'domainaxis1': <DomainAxis: size(10)>,
 'domainaxis2': <DomainAxis: size(9)>}
>>> c.filters_applied()
()
>>> c = c.filter_by_type('dimension_coordinate', 'auxiliary_coordinate')
>>> c.filters_applied()
({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},)
>>> c = c.filter_by_property(units='degrees')
>>> c.filters_applied()
({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},
 {'filter_by_property': ((), {'units': 'degrees'})})
>>> c = c.filter_by_property('or', standard_name='grid_latitude', axis='Y')
>>> c.filters_applied()
({'filter_by_type': ('dimension_coordinate', 'auxiliary_coordinate')},
 {'filter_by_property': ((), {'units': 'degrees'})},
 {'filter_by_property': (('or',), {'axis': 'Y', 'standard_name': 'grid_latitude'})})
>>> print(c)
Constructs:
{'dimensioncoordinate1': <DimensionCoordinate: grid_latitude(10) degrees>}

        '''
        filters = getattr(self, '_filters_applied', None)
        if filters is None:
            return ()

        return deepcopy(filters)
    #--- End: def

    def clear_filters_applied(self):
        '''Remove the history of filters that have been applied.

The removed history is returned in a tuple. The last element of the
tuple describes the last filter applied. Each element is a
single-entry dictionary whose key is the name of the filter method
that was used, with a value that gives the arguments that were passed
to the call of that method. If no filters have been applied then the
tuple is empty.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`, `inverse_filter`,
             `unfilter`

:Returns:

    `tuple`
        The removed history of filters that have been applied, ordered
        from first to last. If no filters have been applied then the
        tuple is empty.


**Examples:**

>>> c.filters_applied()
({'filter_by_naxes': (3, 1)},
 {'filter_by_identity': ('grid_longitude',)})
>>> c.clear_filters_applied()
({'filter_by_naxes': (3, 1)},
 {'filter_by_identity': ('grid_longitude',)})
>>> c.filters_applied()
()

        '''
        out = self.filters_applied()
        self._filters_applied = None
        self._prefiltered = None
        return out
    #--- End: def

    def inverse_filter(self, depth=None):
        '''Return the inverse of previous filters.

By default, the inverse comprises all of the constructs that were
*not* selected by all previously applied filters. If no filters have
been applied, then this will result in empty `Constructs` instance
being returned.

If the *depth* parameter is set to *N* then the inverse is relative to
the constructs selected by the *N*\ -th most recently applied filter.

A history of the filters that have been applied is returned in a
`tuple` by the `filters_applied` method. The last element of the tuple
describes the last filter applied. If no filters have been applied
then the tuple is empty.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `unfilter`

:Parameters:

     depth: `int`, optional
        If set to ``N`` then the inverse is relative to the constructs
        selected by the ``N``-th most recently applied filter. By
        default the inverse is relative to the constructs selected by
        all previously applied filters. ``N`` may be larger than the
        total number of filters applied, which results in the default
        bahaviour.

:Returns:

    `Constructs`
        The constructs, and their construct keys, that were not
        selected by the last filter applied. If no filtering has been
        applied, or the last filter was an inverse filter, then an
        empty `Constructs` instance is returned.

**Examples:**

>>> print(c)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}
>>> print(c.inverse_filter())
Constructs:
{}
>>> d = c.filter_by_type('dimension_coordinate', 'cell_method')
>>> print(d)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >}
>>> print(d.inverse_filter())
Constructs:
{'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}
>>> e = d.filter_by_method('mean')
>>> print(e)
Constructs:
{'cellmethod0': <CellMethod: area: mean>}
>>> print(e.inverse_filter(1))
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >}
>>> print(e.inverse_filter())
Constructs:
{'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}
>>> print(e.inverse_filter(1).inverse_filter())
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}

        '''
        out = self.unfilter(depth=depth) 

        if depth:
            if 'inverse_filter' in self.filters_applied()[-1] :
                filters = out.filters_applied()
                d = 1
                while True:
                    if d > len(filters) or 'inverse_filter' not in filters[-d]:
                        break

                    d += 1

                if d > 1:
                    out = self.unfilter(depth=depth + d - 1)

                return out
        #--- End: if
        
        for key in self:
            out._pop(key)

        out._filters_applied = self.filters_applied() + ({'inverse_filter': ()},)

        out._prefiltered = self.shallow_copy()
        
        return out
    #--- End: def
    
    def shallow_copy(self, _ignore=None):
        '''Return a shallow copy.

``f.shallow_copy()`` is equivalent to ``copy.copy(f)``.

.. versionadded:: 1.7.0

:Returns:

        The shallow copy.

**Examples:**

>>> g = f.shallow_copy()

        '''
        out = super().shallow_copy(_ignore=_ignore)

        prefiltered = getattr(self, '_prefiltered', None)
        if prefiltered is not None:
            out._prefiltered = prefiltered.shallow_copy()
            out._filters_applied = self._filters_applied
           
        return out
    #--- End: def

    def unfilter(self, depth=None):
        '''Return the constructs that existed prior to previous filters.

By default, the unfiltered constructs are those that existed before
all previously applied filters.

If the *depth* parameter is set to *N* then the unfiltered constructs
are those that existed before the *N*\ -th most recently applied
filter.

A history of the filters that have been applied is returned in a
`tuple` by the `filters_applied` method. The last element of the tuple
describes the last filter applied. If no filters have been applied
then the tuple is empty.

.. versionadded:: 1.7.0

.. seealso:: `filter_by_axis`, `filter_by_data`, `filter_by_key`,
             `filter_by_measure`, `filter_by_method`,
             `filter_by_naxes`, `filter_by_identity`,
             `filter_by_ncdim`, `filter_by_ncvar`,
             `filter_by_property`, `filter_by_type`,
             `filters_applied`, `inverse_filter`

:Parameters:

     depth: `int`, optional
        If set to ``N`` then return the constructs selected by the
        ``N``-th most recently applied filter. By default the
        constructs from before all previously applied filters are
        returned. ``N`` may be larger than the total number of filters
        applied, which results in the default bahaviour.

:Returns:

    `Constructs`
        The constructs, and their construct keys, that existed before
        the last filter was applied. If no filters have been applied
        then all of the constructs are returned.

**Examples:**

>>> print(c)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >,
 'domainaxis0': <DomainAxis: size(5)>,
 'domainaxis1': <DomainAxis: size(8)>,
 'domainaxis2': <DomainAxis: size(1)>}
>>> d = c.filter_by_type('dimension_coordinate', 'cell_method')
>>> print(d)
Constructs:
{'cellmethod0': <CellMethod: area: mean>,
 'dimensioncoordinate0': <DimensionCoordinate: latitude(5) degrees_north>,
 'dimensioncoordinate1': <DimensionCoordinate: longitude(8) degrees_east>,
 'dimensioncoordinate2': <DimensionCoordinate: time(1) days since 2018-12-01 >}
>>> d.unfilter().equals(c)
True
>>> e = d.filter_by_method('mean')
>>> print(e)
Constructs:
{'cellmethod0': <CellMethod: area: mean>}
>>> c.unfilter().equals(c)
True
>>> c.unfilter(0).equals(c)
True
>>> e.unfilter().equals(c)
True
>>> e.unfilter(1).equals(d)
True
>>> e.unfilter(2).equals(c)
True

If no filters have been applied then the unfiltered constructs are unchanged:

>>> c.filters_applied()
()
>>> c.unfilter().equals(c)
True

        '''
        out = self
        
        if depth is None:
            while True:
                prefiltered = getattr(out, '_prefiltered', None)
                if prefiltered is None:
                    break
                else:
                    out = prefiltered
        else:
            for _ in range(depth):
                prefiltered = getattr(out, '_prefiltered', None)
                if prefiltered is not None:
                    out = prefiltered
                else:
                    break
        #--- End: if
        
        return out.shallow_copy()
    #--- End: def
    
#--- End: class



from __future__ import print_function

import datetime
import inspect
import os
import unittest

import numpy

import cfdm

class DomainTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        self.test_only = []
    #--- End: def

    def test_DomainAxis__repr__str_construct_type(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        for d in f.domain_axes.values():            
            _ = repr(d)
            _ = str(d)
            self.assertTrue(d.construct_type == 'domain_axis')
    #--- End: def

    def test_DomainAxis_equals(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f

        d = list(f.domain_axes.values())[0]
        e = d.copy()

        self.assertTrue(d.equals(d, verbose=True))
        self.assertTrue(d.equals(e, verbose=True))
        self.assertTrue(e.equals(d, verbose=True))
    #--- End: def

    def test_DomainAxis_unlimited(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = self.f.copy()

        for d in f.domain_axes.values():            
            self.assertFalse(d.nc_is_unlimited())
            d.nc_set_unlimited(False)
            self.assertFalse(d.nc_is_unlimited())
            d.nc_set_unlimited(True)
            self.assertTrue(d.nc_is_unlimited())
            d.nc_set_unlimited(False)
            self.assertFalse(d.nc_is_unlimited())            
    #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print()
    unittest.main(verbosity=2)



from __future__ import print_function

import datetime
import inspect
import os
import tempfile
import unittest

import cfdm

class NetCDFTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')
        f = cfdm.read(self.filename)
        self.assertTrue(len(f)==1, 'f={!r}'.format(f))
        self.f = f[0]

        (fd, self.tempfilename) = tempfile.mkstemp(suffix='.nc', prefix='cfdm_', dir='.')
        os.close(fd)
        
        self.test_only = []
    #--- End: def

    def tearDown(self):
        os.remove(self.tempfilename)
    #--- End: def
    
    def test_netCDF_variable_dimension(self):
        if self.test_only and inspect.stack()[0][3] not in self.test_only:
            return

        f = cfdm.Field()

        f.nc_set_variable('qwerty')
        self.assertTrue(f.nc_has_variable())
        self.assertTrue(f.nc_get_variable() == 'qwerty')
        self.assertTrue(f.nc_get_variable(default=None) == 'qwerty')
        self.assertTrue(f.nc_del_variable() == 'qwerty')
        self.assertFalse(f.nc_has_variable())
        self.assertTrue(f.nc_get_variable(default=None) == None)
        self.assertTrue(f.nc_del_variable(default=None) == None)

        d = cfdm.DomainAxis()
        
        d.nc_set_dimension('qwerty')
        self.assertTrue(d.nc_has_dimension())
        self.assertTrue(d.nc_get_dimension() == 'qwerty')
        self.assertTrue(d.nc_get_dimension(default=None) == 'qwerty')
        self.assertTrue(d.nc_del_dimension() == 'qwerty')
        self.assertFalse(d.nc_has_dimension())
        self.assertTrue(d.nc_get_dimension(default=None) == None)
        self.assertTrue(d.nc_del_dimension(default=None) == None)

        d = cfdm.Count()
        
        d.nc_set_sample_dimension('qwerty')
        self.assertTrue(d.nc_has_sample_dimension())
        self.assertTrue(d.nc_get_sample_dimension() == 'qwerty')
        self.assertTrue(d.nc_get_sample_dimension(default=None) == 'qwerty')
        self.assertTrue(d.nc_del_sample_dimension() == 'qwerty')
        self.assertFalse(d.nc_has_sample_dimension())
        self.assertTrue(d.nc_get_sample_dimension(default=None) == None)
        self.assertTrue(d.nc_del_sample_dimension(default=None) == None)
    #--- End: def


#    def test_netCDF_global_unlimited(self):
#        if self.test_only and inspect.stack()[0][3] not in self.test_only:
#            return
#
#        # ------------------------------------------------------------
#        # Unlimited dimensions
#        # ------------------------------------------------------------
#        f = cfdm.Field()
#        self.assertTrue(f.nc_clear_unlimited_dimensions() == set())
#
#        f = cfdm.Field()
#        f.nc_set_unlimited_dimensions(())
#        
#        f = cfdm.Field()
#        self.assertTrue(f.nc_unlimited_dimensions() == set())
#        f.nc_set_unlimited_dimensions(['qwerty', 'asdf'])
#        self.assertTrue(f.nc_unlimited_dimensions() == set(['qwerty', 'asdf']))
#        f.nc_set_unlimited_dimensions(['zxc'])
#        self.assertTrue(f.nc_unlimited_dimensions() == set(['qwerty', 'asdf', 'zxc']))
#        self.assertTrue(f.nc_clear_unlimited_dimensions() == set(['qwerty', 'asdf', 'zxc']))
#        self.assertTrue(f.nc_unlimited_dimensions() == set())
       
        # ------------------------------------------------------------
        # Global attributes
        # ------------------------------------------------------------
        f = cfdm.Field()
        self.assertTrue(f.nc_clear_global_attributes() == {})
        
        f.nc_set_global_attribute('Conventions')
        f.nc_set_global_attribute('project', 'X')
        self.assertTrue(f.nc_global_attributes() == {'Conventions': None,
                                                     'project': 'X'})
        
        f.nc_set_global_attribute('project')
        f.nc_set_global_attribute('comment', None)
        self.assertTrue(f.nc_global_attributes() == {'Conventions': None,
                                                     'project': None,
                                                     'comment': None})
        
        self.assertTrue(f.nc_clear_global_attributes() == {'Conventions': None,
                                                           'project': None,
                                                           'comment': None})
        self.assertTrue(f.nc_global_attributes() == {})
        
        f.nc_set_global_attribute('Conventions')
        f.nc_set_global_attribute('project')
        self.assertTrue(f.nc_global_attributes() == {'Conventions': None,
                                                     'project': None})


        f = cfdm.Field()
        f.set_properties({'foo': 'bar', 'comment': 'variable comment'})
        f.nc_set_variable('tas')
        d = f.set_construct(cfdm.DomainAxis(2))
        f.set_data(cfdm.Data([8, 9]), axes=[d])

        f2 = f.copy()
        f2.nc_set_variable('ua')
        
        cfdm.write([f, f2], 'tempfilename.nc', file_descriptors={'comment': 'global comment',
                                                                 'qwerty': 'asdf'})

        g = cfdm.read('tempfilename.nc', verbose=False)
        self.assertTrue(len(g) == 2)
        
        for x in g:
            self.assertTrue(x.properties() == {'comment': 'variable comment',
                                               'foo': 'bar',
                                               'qwerty': 'asdf',
                                               'Conventions': 'CF-'+cfdm.CF()})
            self.assertTrue(x.nc_global_attributes() == {'comment': 'global comment',
                                                         'qwerty': None,
                                                         'Conventions': None},
                            x.nc_global_attributes())

        cfdm.write(g, 'tempfilename2.nc')
        h = cfdm.read('tempfilename2.nc')
        for x, y in zip(h, g):
            self.assertTrue(x.properties()           == y.properties())
            self.assertTrue(x.nc_global_attributes() == y.nc_global_attributes())
            self.assertTrue(x.equals(y, verbose=True))
            self.assertTrue(y.equals(x, verbose=True))

        g[1].nc_set_global_attribute('comment', 'different comment')
        cfdm.write(g, 'tempfilename3.nc')
        h = cfdm.read('tempfilename3.nc')
        for x, y in zip(h, g):
            self.assertTrue(x.properties() == y.properties())
            self.assertTrue(x.nc_global_attributes() == {'comment': None,
                                                         'qwerty': None,
                                                         'Conventions': None})
            self.assertTrue(x.equals(y, verbose=True))
            self.assertTrue(y.equals(x, verbose=True))
#        os.remove('tempfilename.nc')
   #--- End: def

#--- End: class

if __name__ == '__main__':
    print('Run date:', datetime.datetime.now())
    print(cfdm.environment(display=False))
    print('')
    unittest.main(verbosity=2)



from __future__ import print_function
from builtins import (str, super, zip)

from . import mixin
from . import core
from . import Constructs
from . import Domain


class Field(mixin.NetCDFVariable,
            mixin.NetCDFGeometry,
            mixin.NetCDFGlobalAttributes,
#            mixin.NetCDFUnlimitedDimensions,
            mixin.ConstructAccess,
            mixin.PropertiesData,
            core.Field):
    '''A field construct of the CF data model.

The field construct is central to the CF data model, and includes all
the other constructs. A field corresponds to a CF-netCDF data variable
with all of its metadata. All CF-netCDF elements are mapped to a field
construct or some element of the CF field construct. The field
construct contains all the data and metadata which can be extracted
from the file using the CF conventions.

The field construct consists of a data array and the definition of its
domain (that describes the locations of each cell of the data array),
field ancillary constructs containing metadata defined over the same
domain, and cell method constructs to describe how the cell values
represent the variation of the physical quantity within the cells of
the domain. The domain is defined collectively by the following
constructs of the CF data model: domain axis, dimension coordinate,
auxiliary coordinate, cell measure, coordinate reference and domain
ancillary constructs.

The field construct also has optional properties to describe aspects
of the data that are independent of the domain. These correspond to
some netCDF attributes of variables (e.g. units, long_name and
standard_name), and some netCDF global file attributes (e.g. history
and institution).

**NetCDF interface**

The netCDF variable name of the construct may be accessed with the
`nc_set_variable`, `nc_get_variable`, `nc_del_variable` and
`nc_has_variable` methods.

.. versionadded:: 1.7.0

    '''
    def __new__(cls, *args, **kwargs):
        '''
        '''
        instance = super().__new__(cls)
        instance._Constructs = Constructs
        instance._Domain     = Domain
        return instance
    #--- End: def

    def __init__(self, properties=None, source=None, copy=True,
                 _use_data=True):
        '''**Initialization**

:Parameters:

    properties: `dict`, optional
        Set descriptive properties. The dictionary keys are property
        names, with corresponding values. Ignored if the *source*
        parameter is set.

        *Parameter example:*
          ``properties={'standard_name': 'air_temperature'}``
        
        Properties may also be set after initialisation with the
        `set_properties` and `set_property` methods.

    source: optional
        Initialize the properties, data and metadata constructs from
        those of *source*.
        
    copy: `bool`, optional
        If False then do not deep copy input parameters prior to
        initialization. By default arguments are deep copied.

        '''        
        # Initialize the new field with attributes and CF properties
        core.Field.__init__(self, properties=properties,
                            source=source, copy=copy,
                            _use_data=_use_data)
        
        self._initialise_netcdf(source)
    #--- End: def

    def __repr__(self):
        '''Called by the `repr` built-in function.

x.__repr__() <==> repr(x)

.. versionadded:: 1.7.0

        '''
        return '<{0}: {1}>'.format(self.__class__.__name__,
                                   self._one_line_description())
    #--- End: def

    def __str__(self):
        '''Called by the `str` built-in function.

x.__str__() <==> str(x)

.. versionadded:: 1.7.0

        '''
        title = "Field: {0}".format(self.identity(''))

        # Append the netCDF variable name
        ncvar = self.nc_get_variable(None)
        if ncvar is not None:
            title += " (ncvar%{0})".format(ncvar)
        
        string = [title]
        string.append(''.ljust(len(string[0]), '-'))

        # Units
        units = getattr(self, 'units', '')
        calendar = getattr(self, 'calendar', None)
        if calendar is not None:
            units += ' {0}'.format(calendar)
            
        # Axes
        data_axes = self.get_data_axes(default=())
        non_spanning_axes = set(self.domain_axes).difference(data_axes)

        axis_names = self._unique_domain_axis_identities()
        
        # Data
        string.append(
            'Data            : {0}'.format(
                self._one_line_description(axis_names)))

        # Cell methods
        cell_methods = self.cell_methods
        if cell_methods:
            x = []
            for cm in cell_methods.values():
                cm = cm.copy()
                cm.set_axes(tuple([axis_names.get(axis, axis)
                                   for axis in cm.get_axes(())]))
                x.append(str(cm))
                
            c = ' '.join(x)
            
            string.append('Cell methods    : {0}'.format(c))
        #--- End: if
        
        def _print_item(self, key, variable, axes):
            '''Private function called by __str__

            '''
            # Field ancillary
            x = [variable.identity(default=key)]

            if variable.has_data():
                shape = [axis_names[axis] for axis in axes]
                shape = str(tuple(shape)).replace("'", "")
                shape = shape.replace(',)', ')')
                x.append(shape)
            elif hasattr(variable, 'nc_get_external'):
                if variable.nc_get_external():
                    ncvar = variable.nc_get_variable(None)
                    if ncvar is not None:
                        x.append(' (external variable: ncvar%{})'.format(ncvar))
                    else:
                        x.append(' (external variable)')
            #--- End: if
                
            if variable.has_data():
                x.append(' = {0}'.format(variable.get_data()))
                
            return ''.join(x)
        #--- End: def
                          
        # Field ancillary variables
        x = [_print_item(self, key, anc, self.constructs.data_axes()[key])
             for key, anc in sorted(self.field_ancillaries.items())]
        if x:
            string.append('Field ancils    : {}'.format(
                '\n                : '.join(x)))


        string.append(str(self.domain))
        
        return '\n'.join(string)
    #--- End def

    def __getitem__(self, indices):
        '''Return a subspace of the field defined by indices.

f.__getitem__(indices) <==> f[indices]

The new subspace contains the same properties and similar metadata
constructs to the original field, but the latter are also subspaced
when they span domain axis constructs that have been changed.

Indexing follows rules that are very similar to the numpy indexing
rules, the only differences being:

* An integer index i takes the i-th element but does not reduce the
  rank by one.

* When two or more dimensions' indices are sequences of integers then
  these indices work independently along each dimension (similar to
  the way vector subscripts work in Fortran). This is the same
  behaviour as indexing on a Variable object of the netCDF4 package.

.. versionadded:: 1.7.0

:Returns:

    `Field`
        The subspace of the field construct.

**Examples:**

>>> f.data.shape
(1, 10, 9)
>>> f[:, :, 1].data.shape
(1, 10, 1)
>>> f[:, 0].data.shape
(1, 1, 9)
>>> f[..., 6:3:-1, 3:6].data.shape
(1, 3, 3)
>>> f[0, [2, 9], [4, 8]].data.shape
(1, 2, 2)
>>> f[0, :, -2].data.shape
(1, 10, 1)

        ''' 
        data  = self.get_data()
        shape = data.shape

        indices = data._parse_indices(indices)
        indices = tuple(indices)
        
        new = self.copy() #data=False)

        data_axes = self.get_data_axes()
        
        # Open any files that contained the original data (this not
        # necessary, is an optimisation)
        
        # ------------------------------------------------------------
        # Subspace the field's data
        # ------------------------------------------------------------
#        new.set_data(data[tuple(indices)], axes=None) #, data_axes)
        new_data = data[tuple(indices)]

        # Replace domain axes
        domain_axes = new.domain_axes
        for key, size in zip(data_axes, new_data.shape):
            domain_axis = domain_axes[key]
            domain_axis.set_size(size)
            new.set_construct(domain_axis, key=key)
        
        # ------------------------------------------------------------
        # Subspace other constructs that contain arrays
        # ------------------------------------------------------------
        self_constructs = self.constructs
        new_constructs_data_axes = new.constructs.data_axes()
        
        if data_axes:
#            for key, construct in new.constructs.filter_by_data().items():
            for key, construct in new.constructs.filter_by_axis('or', *data_axes).items():
                needs_slicing = False
                dice = []
                for axis in new_constructs_data_axes[key]:
                    if axis in data_axes:
                        needs_slicing = True
                        dice.append(indices[data_axes.index(axis)])
                    else:
                        dice.append(slice(None))
                #--- End: for
    
                if needs_slicing:
#                    new_construct = construct[tuple(dice)]
                    new.set_construct(construct[tuple(dice)], key=key, copy=False)
#                else:
#                    new_construct = construct.copy()
#                    
#                new.set_construct(new_construct, key=key, copy=False)
        #--- End: if

#        new.set_data_axes(axes=data_axes)
        new.set_data(new_data, copy=False)
       
        return new
    #--- End: def

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _one_line_description(self, axis_names_sizes=None):
        '''
        '''
        if axis_names_sizes is None:
            axis_names_sizes = self._unique_domain_axis_identities()
            
        x = [axis_names_sizes[axis] for axis in self.get_data_axes(default=())]
        axis_names = ', '.join(x)
        if axis_names:
            axis_names = '({0})'.format(axis_names)
            
        # Field units        
        units    = self.get_property('units', None)
        calendar = self.get_property('calendar', None)
        if units is not None:
            units = ' {0}'.format(units)
        else:
            units = ''
            
        if calendar is not None:
            units += ' {0}'.format(calendar)
            
        return "{0}{1}{2}".format(self.identity(''), axis_names, units)
    #--- End: def

    def _set_dataset_compliance(self, value):
        '''Set the report of problems encountered whilst reading the field
construct from a dataset.

.. versionadded:: 1.7.0

.. seealso:: `dataset_compliance`

:Parameters:

    value:
        The value of the data_compliance component

:Returns:

    `None`

**Examples:**

        '''
        self._set_component('dataset_compliance', value, copy=True)
    #--- End: def    

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def field_ancillaries(self):
        '''Return field ancillary constructs.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    `Constructs`
        The field ancillary constructs and their construct keys.

**Examples:**

>>> print(f.field_ancillaries)
Constructs:
{}

>>> print(f.field_ancillaries)
Constructs:
{'fieldancillary0': <FieldAncillary: air_temperature standard_error(10, 9) K>}

>>> print(f.field_ancillaries('specific_humuidity standard_error'))
Constructs:
{'fieldancillary0': <FieldAncillary: specific_humidity standard_error(10, 9) K>}

        '''
        return self.constructs.filter_by_type('field_ancillary')
    #--- End: def

    @property
    def cell_methods(self):
        '''Return cell method constructs.

The cell methods are not returned in the order in which they were
applied. To achieve this use the `~Constructs.ordered` of the returned
`Constructs` instance.

.. versionadded:: 1.7.0

.. seealso:: `constructs`, `get_construct`, `set_construct`

:Parameters:

    copy: `bool`, optional
        If True then return copies of the constructs. By default the
        constructs are not copied.

:Returns:

    `Constructs`
        The cell method constructs and their construct keys.

**Examples:**

>>> f.cell_methods
Constructs:
{}

>>> f.cell_methods
Constructs:
{'cellmethod1': <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>,
 'cellmethod0': <CellMethod: domainaxis3: maximum>}

>>> f.cell_methods.ordered()
OrderedDict([('cellmethod0', <CellMethod: domainaxis1: domainaxis2: mean where land (interval: 0.1 degrees)>),
             ('cellmethod1', <CellMethod: domainaxis3: maximum>)])

        '''
        return self.constructs.filter_by_type('cell_method')
    #--- End: def

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def climatological_time_axes(self):
        '''TODO

.. versionadded:: 1.7.0

:Returns:

    `list`
        TODO

**Examples:**

TODO

        '''
        out = []
        
        domain_axes = self.domain_axes
        
        for key, cm in self.cell_methods.ordered().items():
            qualifiers = cm.qualifiers()
            if not ('within' in qualifiers or 'over' in qualifiers):
                continue

            axes = cm.get_axes(default=())
            if len(axes) != 1:
                continue        

            axis = axes[0]
            if axis not in domain_axes:
                continue

            # Still here? Then this axis is a climatological time axis
            out.append((axis,))

        return out
    #--- End: def

    def copy(self, data=True):
        '''Return a deep copy of the field construct.

``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

Arrays within `Data` instances are copied with a copy-on-write
technique. This means that a copy takes up very little extra memory,
even when the original contains very large data arrays, and the copy
operation is fast.

.. versionadded:: 1.7.0

:Parameters:

    data: `bool`, optional
        If False then do not copy the data field construct, nor that
        of any of its metadata constructs. By default all data are
        copied.

:Returns:

        The deep copy.

**Examples:**

>>> g = f.copy()
>>> g = f.copy(data=False)
>>> g.has_data()
False

        '''
        new = super().copy(data=data)

        new._set_dataset_compliance(self.dataset_compliance())

        return new
    #--- End: def

    def dump(self, display=True, _level=0, _title=None):
        '''A full description of the field construct.

Returns a description of all properties, including those of metadata
constructs and their components, and provides selected values of all
data arrays.

.. versionadded:: 1.7.0

:Parameters:

    display: `bool`, optional
        If False then return the description as a string. By default
        the description is printed.

:Returns:

    `None` or `str`
        The description. If *display* is True then the description is
        printed and `None` is returned. Otherwise the description is
        returned as a string.

        '''
        indent = '    '      
        indent0 = indent * _level
        indent1 = indent0 + indent

        if _title is None:
            ncvar = self.nc_get_variable(None)
            _title = self.identity(default=None)
            if ncvar is not None:
                if _title is None:
                    _title = "ncvar%{0}".format(ncvar)
                else:
                    _title += " (ncvar%{0})".format(ncvar)
            #--- End: if
            if _title is None:
                _title = ''
                
            _title = 'Field: {0}'.format(_title)
        #--- End: if

        line  = '{0}{1}'.format(indent0, ''.ljust(len(_title), '-'))

        # Title
        string = [line, indent0+_title, line]

        axis_to_name = self._unique_domain_axis_identities()

        name = self._unique_construct_names()

        constructs_data_axes = self.constructs.data_axes()
        
        # Simple properties
        properties = self.properties()
        if properties:
           string.append(self._dump_properties(_level=_level))
               
        # Data
        data = self.get_data(None)
        if data is not None:
            x = [axis_to_name[axis] for axis in self.get_data_axes(default=())]

            units = self.get_property('units', None)
            if units is None:
                isreftime = bool(self.get_property('calendar', False))
            else:
                isreftime = 'since' in units
    
            if isreftime:
                data = data.asdata(data.datatime_array)
                
            string.append('')
            string.append('{0}Data({1}) = {2}'.format(indent0,
                                                      ', '.join(x),
                                                      str(data)))
            string.append('')

        # Cell methods
        cell_methods = self.cell_methods
        if cell_methods:
            for cm in cell_methods.values():
                cm = cm.copy()
                cm.set_axes(tuple([axis_to_name.get(axis, axis)
                                   for axis in cm.get_axes(())]))
                string.append(cm.dump(display=False,  _level=_level))

            string.append('') 
        #--- End: if

        # Field ancillaries        
        for cid, value in sorted(self.field_ancillaries.items()):
            string.append(value.dump(display=False,
                                     _axes=constructs_data_axes[cid],
                                     _axis_names=axis_to_name,
                                     _level=_level))
            string.append('') 

        string.append(self.get_domain().dump(display=False))
        
        string = '\n'.join(string)
       
        if display:
            print(string)
        else:
            return string
    #--- End: def

    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_compression=False,
               ignore_type=False):
        '''Whether two field constructs are the same.

Equality is strict by default. This means that for two field
constructs to be considered equal they must have corresponding
metadata constructs and for each pair of constructs:

* the same descriptive properties must be present, with the same
  values and data types, and vector-valued properties must also have
  same the size and be element-wise equal (see the *ignore_properties*
  and *ignore_data_type* parameters), and

..

* if there are data arrays then they must have same shape and data
  type, the same missing data mask, and be element-wise equal (see the
  *ignore_data_type* parameter).

Two real numbers ``x`` and ``y`` are considered equal if
``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
differences) and ``rtol`` (the tolerance on relative differences) are
positive, typically very small numbers. See the *atol* and *rtol*
parameters.

If data arrays are compressed then the compression type and the
underlying compressed arrays must be the same, as well as the arrays
in their uncompressed forms. See the *ignore_compression* parameter.

Any type of object may be tested but, in general, equality is only
possible with another field construct, or a subclass of one. See the
*ignore_type* parameter.

NetCDF elements, such as netCDF variable and dimension names, do not
constitute part of the CF data model and so are not checked on any
construct.

.. versionadded:: 1.7.0

:Parameters:

    other: 
        The object to compare for equality.

    atol: float, optional
        The tolerance on absolute differences between real
        numbers. The default value is set by the `cfdm.ATOL` function.
        
    rtol: float, optional
        The tolerance on relative differences between real
        numbers. The default value is set by the `cfdm.RTOL` function.

    ignore_fill_value: `bool`, optional
        If True then the "_FillValue" and "missing_value" properties
        are omitted from the comparison, for the field construct and
        metadata constructs.

    verbose: `bool`, optional
        If True then print information about differences that lead to
        inequality.

    ignore_properties: sequence of `str`, optional
        The names of properties of the field construct (not the
        metadata constructs) to omit from the comparison. Note that
        the "Conventions" property is always omitted by default.

    ignore_data_type: `bool`, optional
        If True then ignore the data types in all numerical 
        comparisons. By default different numerical data types imply
        inequality, regardless of whether the elements are within the
        tolerance for equality.

    ignore_compression: `bool`, optional
        If True then any compression applied to underlying arrays is
        ignored and only uncompressed arrays are tested for
        equality. By default the compression type and, if applicable,
        the underlying compressed arrays must be the same, as well as
        the arrays in their uncompressed forms

    ignore_type: `bool`, optional
        Any type of object may be tested but, in general, equality is
        only possible with another field construct, or a subclass of
        one. If *ignore_type* is True then ``Field(source=other)`` is
        tested, rather than the ``other`` defined by the *other*
        parameter.

:Returns: 
  
    `bool`
        Whether the two field constructs are equal.

**Examples:**

>>> f.equals(f)
True
>>> f.equals(f.copy())
True
>>> f.equals(f[...])
True
>>> f.equals('not a Field instance')
False

>>> g = f.copy()
>>> g.set_property('foo', 'bar')
>>> f.equals(g)
False
>>> f.equals(g, verbose=True)
Field: Non-common property name: foo
Field: Different properties
False

        '''
        # ------------------------------------------------------------
        # Check the properties and data
        # ------------------------------------------------------------
        ignore_properties = tuple(ignore_properties) + ('Conventions',)
            
        if not super().equals(
                other,
                rtol=rtol, atol=atol, verbose=verbose,
                ignore_data_type=ignore_data_type,
                ignore_fill_value=ignore_fill_value,
                ignore_properties=ignore_properties,
                ignore_compression=ignore_compression,
                ignore_type=ignore_type):
            return False

        # ------------------------------------------------------------
        # Check the constructs
        # ------------------------------------------------------------
        if not self._equals(self.constructs, other.constructs,
                            rtol=rtol, atol=atol, verbose=verbose,
                            ignore_data_type=ignore_data_type,
                            ignore_fill_value=ignore_fill_value,
                            ignore_compression=ignore_compression,
                            _ignore_type=False):
            if verbose:
                print("{0}: Different metadata constructs".format(
                    self.__class__.__name__))
            return False

        return True
    #--- End: def
        
    def insert_dimension(self, axis, position=0, inplace=False):
        '''Expand the shape of the data array.

Inserts a new size 1 axis, corresponding to an existing domain axis
construct, into the data array.

.. versionadded:: 1.7.0

.. seealso:: `squeeze`, `transpose`

:Parameters:

    axis: `str`
        The construct identifier of the domain axis construct
        corresponding to the inserted axis.

        *Parameter example:*
          ``axis='domainaxis2'``

    position: `int`, optional
        Specify the position that the new axis will have in the data
        array. By default the new axis has position 0, the slowest
        varying position. Negative integers counting from the last
        position are allowed.

        *Parameter example:*
          ``position=2``

        *Parameter example:*
          ``position=-1``

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

    `Field` or `None`
        The new field construct with expanded data axes. If the
        operation was in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.insert_dimension('domainaxis3').data.shape
(1, 96, 73, 19)
>>> f.insert_dimension('domainaxis3', position=3).data.shape
(19, 73, 96, 1)
>>> f.insert_dimension('domainaxis3', position=-1, inplace=True)
(19, 73, 1, 96)
>>> f.data.shape
(19, 73, 1, 96)

        '''
        if inplace:
            f = self
        else:
            f = self.copy()
            
        domain_axis = f.domain_axes.get(axis, None)
        if domain_axis is None:
            raise ValueError("Can't insert non-existent domain axis: {}".format(
                axis))
        
        if domain_axis.get_size() != 1:
            raise ValueError(
"Can't insert an axis of size {}: {!r}".format(domain_axis.get_size(), axis))

        data_axes = f.get_data_axes(default=None)
        if data_axes is not None:
            if axis in data_axes:
                raise ValueError(
                    "Can't insert a duplicate data array axis: {!r}".format(axis))

            data_axes = list(data_axes)
            data_axes.insert(position, axis)

        # Expand the dims in the field's data array
#        f.data.insert_dimension(position, inplace=True)
        super(Field, f).insert_dimension(position, inplace=True)

        if data_axes is not None:
            f.set_data_axes(data_axes)

        if inplace:
            f = None

        return f
    #--- End: def

    def convert(self, key, full_domain=True):
        '''Convert a metadata construct into a new field construct.

The new field construct has the properties and data of the metadata
construct, and domain axis constructs corresponding to the data. By
default it also contains other metadata constructs (such as dimension
coordinate and coordinate reference constructs) that define its
domain.

The `cfdm.read` function allows a field construct to be derived
directly from a netCDF variable that corresponds to a metadata
construct. In this case, the new field construct will have a domain
limited to that which can be inferred from the corresponding netCDF
variable - typically only domain axis and dimension coordinate
constructs. This will usually result in a different field construct to
that created with the `~Field.convert` method.

.. versionadded:: 1.7.0

.. seealso:: `cfdm.read`

:Parameters:

    key: `str` 
        Convert the metadata construct with the given construct key.

    full_domain: `bool`, optional
        If False then do not create a domain, other than domain axis
        constructs, for the new field construct. By default as much of
        the domain as possible is copied to the new field construct.

:Returns:

    `Field`
        The new field construct.

**Examples:**

>>> f = cfdm.read('file.nc')[0]
>>> print(f)
Field: air_temperature (ncvar%ta)
---------------------------------
Data            : air_temperature(atmosphere_hybrid_height_coordinate(1), grid_latitude(10), grid_longitude(9)) K
Cell methods    : grid_latitude(10): grid_longitude(9): mean where land (interval: 0.1 degrees) time(1): maximum
Field ancils    : air_temperature standard_error(grid_latitude(10), grid_longitude(9)) = [[0.76, ..., 0.32]] K
Dimension coords: atmosphere_hybrid_height_coordinate(1) = [1.5]
                : grid_latitude(10) = [2.2, ..., -1.76] degrees
                : grid_longitude(9) = [-4.7, ..., -1.18] degrees
                : time(1) = [2019-01-01 00:00:00]
Auxiliary coords: latitude(grid_latitude(10), grid_longitude(9)) = [[53.941, ..., 50.225]] degrees_N
                : longitude(grid_longitude(9), grid_latitude(10)) = [[2.004, ..., 8.156]] degrees_E
                : long_name:Grid latitude name(grid_latitude(10)) = [--, ..., kappa]
Cell measures   : measure%area(grid_longitude(9), grid_latitude(10)) = [[2391.9657, ..., 2392.6009]] km2
Coord references: atmosphere_hybrid_height_coordinate
                : rotated_latitude_longitude
Domain ancils   : ncvar%a(atmosphere_hybrid_height_coordinate(1)) = [10.0] m
                : ncvar%b(atmosphere_hybrid_height_coordinate(1)) = [20.0]
                : surface_altitude(grid_latitude(10), grid_longitude(9)) = [[0.0, ..., 270.0]] m
>>> x = f.convert('domainancillary2')
>>> print(x)
Field: surface_altitude (ncvar%surface_altitude)
------------------------------------------------
Data            : surface_altitude(grid_latitude(10), grid_longitude(9)) m
Dimension coords: grid_latitude(10) = [2.2, ..., -1.76] degrees
                : grid_longitude(9) = [-4.7, ..., -1.18] degrees
Auxiliary coords: latitude(grid_latitude(10), grid_longitude(9)) = [[53.941, ..., 50.225]] degrees_N
                : longitude(grid_longitude(9), grid_latitude(10)) = [[2.004, ..., 8.156]] degrees_E
                : long_name:Grid latitude name(grid_latitude(10)) = [--, ..., kappa]
Cell measures   : measure%area(grid_longitude(9), grid_latitude(10)) = [[2391.9657, ..., 2392.6009]] km2
Coord references: rotated_latitude_longitude
>>> y = f.convert('domainancillary2', full_domain=False)
>>> print(y)
Field: surface_altitude (ncvar%surface_altitude)
------------------------------------------------
Data            : surface_altitude(grid_latitude(10), grid_longitude(9)) m
		   
        '''
        c = self.constructs.filter_by_key(key).value().copy()
        
        # ------------------------------------------------------------
        # Create a new field with the properties and data from the
        # construct
        # ------------------------------------------------------------
        data = c.del_data()
        
        f = type(self)(source=c, copy=True)
        
        # ------------------------------------------------------------
        # Add domain axes
        # ------------------------------------------------------------
        constructs_data_axes = self.constructs.data_axes()
        data_axes = constructs_data_axes.get(key)
        if data_axes is not None:
            for domain_axis in data_axes:
                f.set_construct(self.domain_axes[domain_axis],
                                key=domain_axis, copy=True)
        #--- End: if

        # ------------------------------------------------------------
        # Set data axes
        # ------------------------------------------------------------
        if data_axes is not None:
            f.set_data(data, axes=data_axes)
            
        # ------------------------------------------------------------
        # Add a more complete domain
        # ------------------------------------------------------------
        if full_domain:
            for ccid, construct in self.constructs.filter_by_type(
                    'dimension_coordinate',
                    'auxiliary_coordinate',
                    'cell_measure').items():
                axes = constructs_data_axes.get(ccid)
                if axes is None:
                    continue
                
                if set(axes).issubset(data_axes):
                    f.set_construct(construct, key=ccid, axes=axes, copy=True)
            #--- End: for
            
            # Add coordinate references which span a subset of the item's
            # axes
            for rcid, ref in self.coordinate_references.items():
                new_coordinates = [
                    ccid
                    for ccid in ref.coordinates()
                    if set(constructs_data_axes[ccid]).issubset(data_axes)]

                if not new_coordinates:
                    continue

                # Still here?
                ok = True
                for ccid in ref.coordinate_conversion.domain_ancillaries().values():
                    axes = constructs_data_axes[ccid]
                    if not set(axes).issubset(data_axes):
                        ok = False
                        break
                #--- End: for

                if ok:
                    ref = ref.copy()
                    ref.clear_coordinates()
                    ref.set_coordinates(new_coordinates)
                    f.set_construct(ref, key=rcid, copy=False)

                    # Copy domain ancillary constructs
                    for dakey in ref.coordinate_conversion.domain_ancillaries().values():
                        construct = self.constructs.get(dakey)
                        if construct is not None:
                            axes = constructs_data_axes.get(dakey)                        
                            f.set_construct(construct, key=dakey, axes=axes, copy=True)
            #--- End: for
        #--- End: if
              
        return f
    #--- End: def
   
    def dataset_compliance(self, display=False):
        '''A report of problems encountered whilst reading the field construct
from a dataset.

If the dataset is partially CF-compliant to the extent that it is not
possible to unambiguously map an element of the netCDF dataset to an
element of the CF data model, then a field construct is still returned
by the `read` function, but may be incomplete.

Such "structural" non-compliance would occur, for example, if the
"coordinates" attribute of a CF-netCDF data variable refers to another
variable that does not exist, or refers to a variable that spans a
netCDF dimension that does not apply to the data variable.

Other types of non-compliance are not checked, such whether or not
controlled vocabularies have been adhered to.

.. versionadded:: 1.7.0

.. seealso:: `cfdm.read`

:Parameters:

    display: `bool`, optional
        If True print the compliance report. By default the report is
        returned as a dictionary.

:Returns:

    `None` or `dict`
        The report. If *display* is True then the report is printed
        and `None` is returned. Otherwise the report is returned as a
        dictionary.

**Examples:**

If no problems were encountered, an empty dictionary is returned:

>>> f.dataset_compliance()
{}

        '''
        d = self._get_component('dataset_compliance', {})

        if not display:
            return d
        
        if not d:
            print(d)
            return
    
        for key0, value0 in d.items():
            print('{{{0!r}:'.format(key0))
            print('    CF version: {0!r},'.format(value0['CF version']))
            print('    dimensions: {0!r},'.format(value0['dimensions']))
            print('    non-compliance: {')
            for key1, value1 in sorted(value0['non-compliance'].items()):
                for x in value1:
                    print('        {!r}: ['.format(key1))
                    print('            {{{0}}},'.format(
                        '\n             '.join(['{0!r}: {1!r},'.format(key2, value2)
                                                for key2, value2 in sorted(x.items())])))
                #--- End: for
                print('        ],')

            print('    },')
            print('}\n')
    #--- End: def
     
    def squeeze(self, axes=None, inplace=False):
        '''Remove size one axes from the data array.

By default all size one axes are removed, but particular size one axes
may be selected for removal.

.. versionadded:: 1.7.0

.. seealso:: `insert_dimension`, `transpose`

:Parameters:

    axes: (sequence of) `int`, optional
        The positions of the size one axes to be removed. By default
        all size one axes are removed. Each axis is identified by its
        original integer position. Negative integers counting from the
        last position are allowed.

        *Parameter example:*
          ``axes=0``

        *Parameter example:*
          ``axes=-2``

        *Parameter example:*
          ``axes=[2, 0]``

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

    `Field` or `None`
        The field construct with removed data axes. If the operation
        was in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(1, 73, 1, 96)
>>> f.squeeze().data.shape
(73, 96)
>>> f.squeeze(0).data.shape
(73, 1, 96)
>>> f.squeeze([-3, 2], inplace=True)
>>> f.data.shape
(73, 96)

        '''
        if inplace:            
            f = self
        else:
            f = self.copy()
            
        if axes is None:
            iaxes = [i for i, n in enumerate(f.data.shape) if n == 1]
        else:
            try:
                iaxes = f.data._parse_axes(axes)
            except ValueError as error:
                raise ValueError("Can't squeeze data: {}".format(error))

        data_axes = f.get_data_axes(default=None)
        if data_axes is  not None:
            new_data_axes = [data_axes[i]
                             for i in range(f.data.ndim) if i not in iaxes]
            
        # Squeeze the field's data array
#        f.data.squeeze(iaxes, inplace=True)
        super(Field, f).squeeze(iaxes, inplace=True)

        if data_axes is not None:
            f.set_data_axes(new_data_axes)

        if inplace:
            f = None

        return f
    #--- End: def

    def transpose(self, axes=None, constructs=False, inplace=False):
        '''Permute the axes of the data array.

.. versionadded:: 1.7.0

.. seealso:: `insert_dimension`, `squeeze`

:Parameters:

    axes: (sequence of) `int`
        The new axis order. By default the order is reversed. Each
        axis in the new order is identified by its original integer
        position. Negative integers counting from the last position
        are allowed.

        *Parameter example:*
          ``axes=[2, 0, 1]``

        *Parameter example:*
          ``axes=[-1, 0, 1]``

    constructs: `bool`
        If True then tranpose the metadata constructs to have the same
        relative domain axis order as the data of tranposed field
        constuct. By default, metadata constructs are not changed.

    inplace: `bool`, optional
        If True then do the operation in-place and return `None`.

:Returns:

    `Field` or `None`
        The field construct with permuted data axes. If the operation
        was in-place then `None` is returned.

**Examples:**

>>> f.data.shape
(19, 73, 96)
>>> f.transpose().data.shape
(96, 73, 19)
>>> f.transpose([1, 0, 2]).data.shape
(73, 19, 96)
>>> f.transpose(inplace=True)
>>> f.data.shape
(96, 19, 73)

        '''
        if inplace:
            f = self
        else:
            f = self.copy()
            
        try:
            iaxes = f.data._parse_axes(axes)
        except ValueError as error:
            raise ValueError("Can't transpose data: {}".format(error))

        ndim = f.data.ndim        
        if iaxes is None:
            iaxes = tuple(range(ndim-1, -1, -1))

        data_axes = f.get_data_axes(default=None)

        # Transpose the field's data array
        super(Field, f).transpose(iaxes, inplace=True)

        if data_axes is not None:
            new_data_axes = [data_axes[i] for i in iaxes]
            f.set_data_axes(new_data_axes)

        if constructs:
            for key, construct in f.constructs.filter_by_data().items():
                data = construct.get_data(None)
                if data is None:
                    continue
                
                if data.ndim < 2:
                    # No need to transpose 1-d constructs
                    continue
                    
                construct_axes = f.get_data_axes(key)

                new_construct_axes = [axis for axis in new_data_axes
                                      if axis in construct_axes]

                for i, axis in enumerate(construct_axes):
                    if axis not in new_construct_axes:
                        new_construct_axes.insert(i, axis)
                #--- End: for
                
                iaxes = [construct_axes.index(axis) for axis in new_construct_axes]
    
                # Transpose the construct
                construct.transpose(iaxes, inplace=True)
    
                f.set_data_axes(axes=new_construct_axes, key=key)
            #--- End: for
        #--- End: if

        if inplace:
            f = None
        
        return f
    #--- End: def

#--- End: class
